/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1988 Electrotechnical Laboratry (ETL)

Permission to use, copy, modify, and distribute this material 
for any purpose and without fee is hereby granted, provided 
that the above copyright notice and this permission notice 
appear in all copies, and that the name of ETL not be 
used in advertising or publicity pertaining to this 
material without the specific, prior written permission 
of an authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY 
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", 
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	vinnntp.c (NNTP interface for VIN)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	940210	reworte to cope with multiple couples of nntp/xnntp servers
///////////////////////////////////////////////////////////////////////*/
#include "conf.h"
#include "nntp.h"
#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
FILE *pseudo_fopen();
typedef int (*IFUNCP)();
IFUNCP set_sigpipe();

typedef unsigned char Uchar;
#define MAX_STATLN	0x4000
#define MAX_COMLN	512
#define MAX_BODYLN	512
#define MAX_LOGLN	10

extern int PREFER_NNTP_HEAD;
	/* set this if nntp HEAD is more effective than direct access */
extern int NNTP_DOPREFETCH;	/* do prefetch */
extern int NNTP_XMIME_ENABLE;	/* send "XMIME enable" on invocation */

char  *NNTPPORT =  "nntp";	/* port name of nntp */
char *XNNTPPORT = "xnntp";	/* port name of xnntp */

int WITH_XNNTP;

#define DONT_RETRY -1

/*
 *	cache of response body
 */
int XNNTPLIB_BANKSIZES;
#define NNTP_respSIZE_INIT	0x10000
#define NNTP_respSIZE_INC	0x4000

typedef struct {
	int	 sv_stat;
	char	*sv_name;
	int	 sv_disable;

	struct {
	int
		without_XHIST:1,
		without_XHEAD:1,
		XNNTP_exists:1;
	} sv_flags;
	FILE	*sv_XNNTP_SERVER[2];

	int	 sv_ndisconnected;
	int	 sv_disconnected;
	int	 sv_resp_pending;
	int	 sv_lastreq_time;

	/* statistics */
	int	 sv_cached;		/* cache hit and reused */
	int	 sv_prefetched;		/* count of prefetching */
	int	 sv_hit;
	int	 sv_miss;

	char	 sv_current_group[MAX_COMLN];
	char	 sv_prefetch_command[MAX_COMLN];
	int	 sv_contiguous;

		/* cache buffer for response body */
	char	 sv_respCOMMAND[MAX_COMLN];	/* command for this data */
	int	 sv_respSIZE;
	Uchar	*sv_respBUFF;			/* response data buffer	*/
	int	 sv_respLENG;			/* response data size	*/

	char	 sv_lastput[MAX_COMLN];
	int	 sv_LOG[MAX_LOGLN][256];
	int	 sv_LOGN;
	int	 sv_LOGX;
} Server;

Server NNTP_SERVERS[4];
static int CUR_SERV;

#define NNTP_STAT		(NNTP_SERVERS[CUR_SERV].sv_stat)
#define NNTP_SERVERNAME		(NNTP_SERVERS[CUR_SERV].sv_name)
#define NNTP_DISABLE		(NNTP_SERVERS[CUR_SERV].sv_disable)

#define NNTP_WITHOUT_XHIST	(NNTP_SERVERS[CUR_SERV].sv_flags.without_XHIST)
#define NNTP_WITHOUT_XHEAD	(NNTP_SERVERS[CUR_SERV].sv_flags.without_XHEAD)
#define NNTP_XNNTP_EXISTS	(NNTP_SERVERS[CUR_SERV].sv_flags.XNNTP_exists)
#define NNTP_XNNTP_SERVER	(NNTP_SERVERS[CUR_SERV].sv_XNNTP_SERVER)

#define NNTP_NDISCONNECTED	(NNTP_SERVERS[CUR_SERV].sv_ndisconnected)
#define NNTP_DISCONNECTED	(NNTP_SERVERS[CUR_SERV].sv_disconnected)
#define NNTP_RESP_PENDING	(NNTP_SERVERS[CUR_SERV].sv_resp_pending)
#define NNTP_LASTREQUEST_TIME	(NNTP_SERVERS[CUR_SERV].sv_lastreq_time)


#define NNTP_CACHED		(NNTP_SERVERS[CUR_SERV].sv_cached)
#define NNTP_PREF		(NNTP_SERVERS[CUR_SERV].sv_prefetched)
#define NNTP_HIT		(NNTP_SERVERS[CUR_SERV].sv_hit)
#define NNTP_MISS		(NNTP_SERVERS[CUR_SERV].sv_miss)

#define NNTP_current_group	(NNTP_SERVERS[CUR_SERV].sv_current_group)
#define NNTP_prefetch_command	(NNTP_SERVERS[CUR_SERV].sv_prefetch_command)
#define NNTP_contiguous		(NNTP_SERVERS[CUR_SERV].sv_contiguous)

#define NNTP_respCOMMAND	(NNTP_SERVERS[CUR_SERV].sv_respCOMMAND)
#define NNTP_respSIZE		(NNTP_SERVERS[CUR_SERV].sv_respSIZE)
#define NNTP_respBUFF		(NNTP_SERVERS[CUR_SERV].sv_respBUFF)
#define NNTP_respLENG		(NNTP_SERVERS[CUR_SERV].sv_respLENG)

#define NNTP_LOG		(NNTP_SERVERS[CUR_SERV].sv_LOG)
#define NNTP_LOGN		(NNTP_SERVERS[CUR_SERV].sv_LOGN)
#define NNTP_LOGX		(NNTP_SERVERS[CUR_SERV].sv_LOGX)
#define NNTP_LastPut		(NNTP_SERVERS[CUR_SERV].sv_lastput)


static jmp_buf nntpenv;
static int *cur_nntpenv;

static sigpipe(){
	NNTP_NDISCONNECTED++;
	if( 1 < NNTP_NDISCONNECTED ){
		errmessage(" FATAL ERROR | nntpd disconnected !",
			NNTP_NDISCONNECTED);
		bell("b1b4");
		NNTP_STAT = DONT_RETRY;
	}else{
		errmessage(" ERROR | nntpd disconnected ?");
		NNTP_DISCONNECTED = 1;
	}
	if( cur_nntpenv )
		longjmp(cur_nntpenv,-1);
}

static addlog(str){
	if( MAX_LOGLN <= ++NNTP_LOGX )
		NNTP_LOGX = 0;
	NNTP_LOGN++;
	StrNcpy(NNTP_LOG[NNTP_LOGX],str,255);
}
static dumplog(){
	char buf[0x8000];
	nntp_getlog(buf);
	fprintf(stderr,"%s",buf);
}
static nntp_getlog(out)
	char *out;
{	int li,ln,ci,ch;
	char *op;

	ln = NNTP_LOGN-MAX_LOGLN;
	*out = 0;
	for(li = NNTP_LOGX+1; ;li++){
		if( MAX_LOGLN <= li )
			li = 0;
		ln++;
		if( NNTP_LOG[li][0] ){
			sprintf(out+strlen(out),"[%3d] ",ln);
			op = out+strlen(out);

			for(ci = 0; ch = NNTP_LOG[li][ci]; ci++){
				if( (ch < 0x20 || 0x7F <= ch) && ch != 033 )
					sprintf(op,"\\%o",ch);
				else	sprintf(op,"%c",ch);
				op += strlen(op);
			}
			sprintf(op,"\n");
		}
		if( ln == NNTP_LOGN )
			break;
	}
}



static PUT_SERVER(command)
	char *command;
{	IFUNCP osig;

	if( NNTP_RESP_PENDING ){
		message(" NNTP | discarding previous response ...");
		nntp_finalize();
		if( nntp_initialize() < 0 )
			return -1;
		NNTP_RESP_PENDING = 0;
		message(" NNTP | discarding previous response | DONE.");
	}

	if( NNTP_STAT <= 0 || NNTP_DISCONNECTED )
		return -1;

	osig = set_sigpipe(sigpipe);
	cur_nntpenv = (int*)nntpenv;
	if( _setjmp(cur_nntpenv) == 0 ){
		strcpy(NNTP_LastPut,command);
		put_server(command);
		addlog(command);
	}
	cur_nntpenv = 0;
	set_sigpipe(osig);

	if( NNTP_DISCONNECTED ){
		dbg_printf("NNTP: ERROR (disconnected)\n");
		nntp_finalize();
		return -1;
	}

	NNTP_LASTREQUEST_TIME = time(0);
	dbg_printf("NNTP.PUT_SERVER: %s\n",command);
	return 0;
}

static GET_SERVER(response,size)
	char *response;
{	int retcode;

	retcode = get_server(response,size);
	addlog(response);
	if( retcode < 0 ){
		fprintf(stderr,"#### NNTP.GET_SERVER failed ####\n");
		fprintf(stderr,"command: %s\n",NNTP_LastPut);
		dumplog();
		fprintf(stderr,"################################\n");
		bell("b1b4");
		sprintf(response,"500 NNTP GET_SERVER FAILED!! (;_;)");
	}
	return retcode;
}
NNTP_GET_SERVER(response,size){ return GET_SERVER(response,size); }

nntp_stat(){
	printf("PREFETCH: %d\n",NNTP_PREF);
	printf("     HIT/ %d\n",NNTP_HIT);
	printf("    MISS/ %d\n",NNTP_MISS);
	printf("  CACHED: %d\n",NNTP_CACHED);

	if( NNTP_prefetch_command[0] )
		printf("PREFETCH-COMMAND: %s\n",NNTP_prefetch_command);
}

static nntp_idle()
{
	if( NNTP_LASTREQUEST_TIME )
		return time(0) - NNTP_LASTREQUEST_TIME;
	else	return -1;
}

char *getenv();
char *getserverbyfile();

static char *
nntp_server(name) char *name; {
	if( NNTP_SERVERNAME == 0 ){
		if( name == 0 ){
			NNTP_SERVERNAME = getenv("NEWS_HOSTNAME");

			if( NNTP_SERVERNAME == 0 )
			if( NNTP_SERVERNAME = getserverbyfile(SERVER_FILE) )
			if(Getenv("NNTPSERVER") == 0)
				message(" DEFAULT NNTPSERVER=%s(%s)",
					NNTP_SERVERNAME,SERVER_FILE);
		}else{
			NNTP_SERVERNAME = (char*)STRING_ALLOC(name);
		}
		if( NNTP_SERVERNAME )
			nntp_server_init0(NNTP_SERVERNAME,NNTPPORT);
	}
	return NNTP_SERVERNAME;
}

static get_version(response,ver)
	char *response,*ver;
{	char *vp;

	*ver = 0;
	for( vp = response; *vp; vp++ ){
		if( strncmp("version",vp,7) == 0 ){
			strcpy(ver,vp);
			return;
		}
	}
}
static ver_message(which,host,response)
{	char ver[256];

	get_version(response,ver);
	message(" [%s] | connected to %s server at %s | %s\n",
		which,which,host,ver);
}

static nntp_initialize(){
	int retcode;
	char response[256];
	char *hostname;

	if( NNTP_STAT < 0 )
		PREFER_NNTP_HEAD = 0;

	if( NNTP_STAT )
		return NNTP_STAT;

	hostname = nntp_server(0);
	if(hostname == NULL){
		/* errmessage(" UNDEF | NNTPSERVER enviroment variable ?"); */
		return NNTP_STAT = DONT_RETRY;
	}
	retcode = nntp_server_init1(hostname,NNTPPORT);
	if(retcode < 0){
		errmessage(" ERROR(%d) | Couldn't connect to NNTP-server",
			retcode);
		return NNTP_STAT = DONT_RETRY;
	}

	get_server(response, sizeof(response));
	retcode = atoi(response);

	if(handle_server_response(retcode,hostname) < 0){
		if( retcode == ERR_GOODBYE )
		     fprintf(stderr,
		     "NNTP FAILED ;-< (maybe the news server is too busy)\n");
		else errmessage(
		     " ERROR | Unknown response(%d) from NNTP-server");
		return NNTP_STAT = DONT_RETRY;
	}
	ver_message("NNTP",hostname,response);
	dbg_printf("NNTP: initialized. server=%s\n",hostname);
	NNTP_STAT = 1;

	nntp_setupcommands();

	return NNTP_STAT;
}
static
nntp_connect()
{
	return nntp_initialize();
}


static
nntp_setupcommands(){
	char resp[MAX_STATLN];

	if( NNTP_XMIME_ENABLE || getenv("NNTP_XMIME_ENABLE") )
	if( PUT_SERVER("XMIME enable") == 0 )
	if( nntp_get_statresp(resp,sizeof(resp),"XMIME") == 0 )
		message(" [NNTP+MIME] | %s |  %s",nntp_server(0),resp);

	NNTP_WITHOUT_XHEAD = 1;

	if( WITH_XNNTP ){
		char *host;

		if( (host = getenv("XNNTPSERVER")) == 0 )
			host = nntp_server(0);
		xnntp_initialize(host);
	}
}
static xnntp_initialize(host){
	if(NNTP_XNNTP_SERVER[0])
		return;

	if( server_Init(NNTP_XNNTP_SERVER,host,XNNTPPORT) < 0 ){
		NNTP_XNNTP_EXISTS = 0;
		fprintf(stderr,"cannot connect XNNTP server at '%s'\n",host);
/*
		message(" XNNTP | cannot connect XNNTP server at %s",host);
		bell("b2");
*/
	}else{
		char xutmp[256],request[256],response[256];

		get_server1(NNTP_XNNTP_SERVER,response,sizeof(response));
		ver_message("XNNTP",host,response);
		NNTP_XNNTP_EXISTS = 1;

/*
#ifdef XNNTP_UTMP
*/
		sprint_xutmp(xutmp,"vin");
		sprintf(request,"XUSER %s",xutmp);
		put_server1(NNTP_XNNTP_SERVER,request);
		get_server1(NNTP_XNNTP_SERVER,response,sizeof(response));
/*
#endif
*/
	}
}

nntp_finalize(){
	if( 0 < NNTP_STAT ){
		NNTP_current_group[0] = 0;
		if( NNTP_RESP_PENDING ){
			NNTP_prefetch_command[0] = 0;
		}else{
			nntp_cancell();
			PUT_SERVER("QUIT");
		}
		NNTP_STAT = 0;
	}
	if( NNTP_DISCONNECTED ){
		nntp_server_done(0);
		NNTP_DISCONNECTED = 0;
		errmessage(" FATAL ERROR | nntpd disconnected ??");
	}else{
		IFUNCP osig;

		osig = set_sigpipe(sigpipe);
		cur_nntpenv = (int*)nntpenv;
		if( _setjmp(cur_nntpenv) == 0 )
			nntp_server_done(1);
		cur_nntpenv = 0;
		set_sigpipe(osig);
	}
	NNTP_LASTREQUEST_TIME = 0;
}

static nntp_set_group( newsgroup, artnumber )
	char *newsgroup;
{	char nntp_combuf[MAX_COMLN];
	char response[MAX_STATLN];

	if(newsgroup == 0 || *newsgroup == 0 )
		return 0;

	if( strcmp(newsgroup,NNTP_current_group) == 0 ){
		NNTP_contiguous++;
		return 1;
	}
	NNTP_current_group[0] = 0;
	NNTP_contiguous = 1;
	nntp_cancell();

	sprintf(nntp_combuf, "GROUP %s", newsgroup);
	if( PUT_SERVER(nntp_combuf) != 0 )
		return 0;

	if( nntp_get_statresp(response,sizeof(response),nntp_combuf) < 0 )
		return 0;

	strcpy(NNTP_current_group,newsgroup);
	return 1;
}

#define noxhist "\n\
XHIST is not built-in in nntpd at [%s].\n\
rsh %s XHIST will be substituted for ...\n"

static nntp_get_statresp(aresp,arespsize,comm)
	register char *aresp,*comm;
{	char *hostname;
	char *resp,respbuf[MAX_STATLN];
		/* must be large enough to receive any long
		   status especially the result of XHIST */

	resp = respbuf;
	*respbuf = 0;
	if( GET_SERVER(respbuf,sizeof(respbuf)) < 0 )
		return -1;

	StrNcpy(aresp,respbuf,arespsize);

	nntp_printlog(0,"> %s",respbuf);
	dbg_printf("NNTP: response %s\n",*resp?resp:comm);


	switch( *resp ){
		case CHAR_INF:
			tcaprintf("\r%<so>NNTP-INFO: %s%<se>",resp);
			message(" NNTP info | %s",resp);
			bell("b4");
			return nntp_get_statresp(aresp,arespsize,comm);

		case CHAR_OK:
		case CHAR_CONT:
			return 0;
		case CHAR_ERR:
			dbg_printf("NNTP ERROR: %s\n",resp);
			break;
		case CHAR_FATAL:
			if( strncmp(comm,"XUSER",5) == 0 ){
				/* xnntp-1.X does not have XUSER */
			}else
			if( strncmp(comm,"XHIST",5) == 0 ){
				hostname = nntp_server(0);
				fprintf(stderr,noxhist,hostname,hostname);
				NNTP_WITHOUT_XHIST = 1;
			}else
			if( strncmp(comm,"XHEAD",5) == 0
			 || strncmp(comm,"XMIME",5) == 0 ){
				NNTP_WITHOUT_XHEAD = 1;
			}else{
				nntp_printlog(0,"> (FATAL) [%s] %s",comm,resp);

fprintf(stderr,"\rNNTP fatail error (%s)(%s)\n",resp,comm);
errmessage(" NNTP fatal error | %s | %s",resp,comm);

				bell("b1b1b4");
			}
	}
	return -1;
}

extern int NNTP_prefetching;
static nntp_prefetch( newsgroup, number,command )
	char *newsgroup;
{	char nntp_combuf[MAX_COMLN];

	if( nntp_initialize() < 0 )
		return;
	if( NNTP_prefetch_command[0] )
		return;
	if( !nntp_set_group(newsgroup,number) )
		return;

	sprintf(nntp_combuf,"%s %ld",command,number);

	NNTP_prefetching = 1;
	if( PUT_SERVER(nntp_combuf) != 0 )
		return;

	strcpy(NNTP_prefetch_command,nntp_combuf);
	NNTP_PREF++;
}

/*##############################################################*
 *	cache buffer for response body				*
 *##############################################################*/
static nntp_inflush(){
	int nline = 0;
	char response[MAX_STATLN];

	while( 0 <= GET_SERVER( response, sizeof(response)) ){
		nline++;
		if ( response[0] == '.' && response[1] == 0 )
			break;
		if((nline & 0xFF)== 0 )
		if( get_busy_count(2) )
		message(" NNTP heavy (-_-;; | discarding ... | %d lines",nline);
	}
}

static FILE *
nntp_getresponse1( xnntp, command, skip, linesp, bytesp )
	char *command;
	int *linesp,*bytesp;
{	register Uchar *bp,*buffer;
	register FILE *fp;
	register int nline,leng,rsize;

	if( skip ){
		nntp_inflush();
		return NULL;
	}

	if(NNTP_respBUFF == 0){
		NNTP_respSIZE = NNTP_respSIZE_INIT;
		NNTP_respBUFF = (Uchar*)allocOBJECT("nntp_getresponse1",
					NNTP_respSIZE);
		XNNTPLIB_BANKSIZES += NNTP_respSIZE;
	}
	fp = pseudo_fopen(NNTP_respBUFF,NNTP_respSIZE);
	bp = buffer = (Uchar*)pseudo_ioptr(fp);

	nline = 0;
	rsize = NNTP_respSIZE;

	NNTP_respCOMMAND[0] = 0;
	NNTP_RESP_PENDING = 1;

	for(;;){
		if( xnntp ){
			if( get_server1(NNTP_XNNTP_SERVER,bp,rsize) < 0 ) break;
		}else	if( GET_SERVER(bp,rsize) < 0 ) break;

		if ( bp[0] == '.' ){
			if( bp[1] == 0 )
				break;
			if( bp[1] == '.' )
				strcpy(bp,&bp[1]);
		}

		leng = strlen(bp);
		bp += leng;
		*bp++ = '\n';
		nline++;
		rsize -= (leng + 1);

		if( rsize <= MAX_BODYLN ){
			XNNTPLIB_BANKSIZES += NNTP_respSIZE_INC;
			NNTP_respSIZE += NNTP_respSIZE_INC;
			dbg_printf("NNTP: expand-buffer %d\n",NNTP_respSIZE);
			NNTP_respBUFF = (Uchar*)
					realloc(NNTP_respBUFF,NNTP_respSIZE);
			fclose(fp);
			fp = pseudo_fopen(NNTP_respBUFF,NNTP_respSIZE);
			bp = &NNTP_respBUFF[bp-buffer];
			buffer = NNTP_respBUFF;
			rsize = NNTP_respSIZE_INC;
		}
		if((nline & 0xFF)== 0 )
		if( get_busy_count(2) )
			message(" NNTP heavy (-_-;; | %d bytes/ %d lines | %s",
				bp-buffer,nline,command);
	}
	NNTP_RESP_PENDING = 0;
	*bp = 0;

	leng = bp - buffer;
	pseudo_ungetv(fp,leng);

	if( bytesp ) *bytesp = leng;
	if( linesp ) *linesp = nline;

	NNTP_respLENG = leng;
	strcpy(NNTP_respCOMMAND,command);

	return fp;
}
FILE *
nntp_getresponse( command, skip, linesp, bytesp )
{	char response[MAX_STATLN];

	if( nntp_get_statresp(response,sizeof(response),command) < 0 )
		return NULL;
	return nntp_getresponse1( 0, command, skip, linesp, bytesp );
}

/*
 *	cancell(skip) the result of prefetched command
 */
static nntp_cancell(){
	int anum;

	/* 90.10.24 CLEAR CACHE FOR A CANCELLD ARTICLE */
	NNTP_respCOMMAND[0] = 0;
	if( NNTP_respBUFF )
		NNTP_respBUFF[0] = 0;

	if( NNTP_prefetch_command[0] ){
		NNTP_MISS++;
		nntp_getresponse(NNTP_prefetch_command,1,0,0);
		NNTP_prefetch_command[0] = 0;
	}
}

extern char XHIST_LASTLINE_COMMAND[];
static activedate(){
	int date;
	char line[MAX_STATLN];

	*line = 0;
	nntp_midscan(line,XHIST_LASTLINE_COMMAND);
	if( date = decomp_history_line(line,0,0) )
		return date;

	return time(0);
}

static FILE *
nntp_getactive(hostname,difference,adatep)
	int *adatep;
{	FILE *fp;
	char request[MAX_COMLN],response[MAX_STATLN];
	int date;

	/*################################################################
	 * periodical (and differentical) XLIST/XNNTP may be the only NNTP
	 * job during a Mail User Agent is idle long time.
	 */
	if( NNTP_XNNTP_EXISTS ){
		sprintf(request,"XLIST");
		if( difference ) strcat(request," difference");

		dbg_printf("XNNTP: %s\n",request);
		put_server1(NNTP_XNNTP_SERVER,request);
		get_server1(NNTP_XNNTP_SERVER,response,sizeof(response));
		if( *response == CHAR_OK ){
			if( adatep )
				sscanf(response,"%*d %d",adatep);

			fp = NNTP_XNNTP_SERVER[0];
			dbg_printf("NNTP: %s = %x.\n",request,fp);
			return fp;
		}
	}
	/*################################################################
	 * This should be done after XLIST/XNNTP trial to avoid restarting
         * wasteful connection to normal NNTP
	 */
	if( nntp_initialize() < 0 )
		return NULL;

	if( adatep )
		date = activedate();

	if( PUT_SERVER("LIST") != 0 ){
		fprintf(stderr,"Cannot send LIST command to NNTP server.\n");
		return NULL;
	}
	nntp_cancell();
	if(nntp_get_statresp(response,sizeof(response),"LIST") < 0)
		return NULL;

	fp = (FILE*)nntp_server_fp();
	dbg_printf("NNTP: LIST = %x.\n",fp);
	if( adatep )
		*adatep = date;
	return fp;
}

static FILE *
nntp_newgroups(date,distributions)
	char *distributions;
{	char sdate[MAX_COMLN],command[MAX_COMLN];
	FILE *fp;

	if( nntp_initialize() < 0 )
		return NULL;
	DateYYMMDD_HHMMSS(date,sdate,0,"%s %s");
	sprintf(command,"NEWGROUPS %s %s",sdate,distributions);
	if( PUT_SERVER(command) != 0 )
		return NULL;

	nntp_cancell();
	fp = nntp_getresponse(command,0,0,0);
	if( fp != NULL )
		dbg_printf("NNTP: NEWGROUPS %d bytes.\n",pseudo_fsize(fp));
	return fp;
}

static FILE *nntp_newhistory(date){
	char command[MAX_COMLN];
	FILE *fp;

	if( NNTP_WITHOUT_XHIST )
		return NULL;

	if( nntp_initialize() < 0 )
		return NULL;

	sprintf(command,"XHIST %s",date);
	if( PUT_SERVER(command) != 0 )
		return NULL;

	nntp_cancell();
	fp = nntp_getresponse(command,0,0,0);
	if( fp != NULL ){
		dbg_printf("NNTP: XHIST %d bytes.\n",pseudo_fsize(fp));
		return fp;
	}
	return NULL;
}

/*##############################################################*
 *	getting HEAD, BODY or ARTICLE				*
 *##############################################################*/
static FILE *
nntp_article( command, newsgroup, number, lines, bytes )
	char *command, *newsgroup;
	int *lines, *bytes;
{	char nntp_combuf[MAX_COMLN];
	FILE *fp;

	if( NNTP_DISABLE || nntp_initialize() < 0 )
		return NULL;

	if( !nntp_set_group(newsgroup,number) ){
		errmessage(" ERROR on fetch | %s:%d",newsgroup,number);
		return NULL;
	}

if( !NNTP_WITHOUT_XHEAD && strcmp(command,"HEAD")==0 )
command = "XHEAD";

	sprintf(nntp_combuf,"%s %ld",command,number);
	/*########################################################
	 *	reuse prevoious HEAD, ARTICLE and BODY if possible
	 */
	if( strcmp(NNTP_respCOMMAND,nntp_combuf) == 0 ){
		fp = pseudo_fopen(NNTP_respBUFF,NNTP_respSIZE);
		pseudo_ungetv(fp,NNTP_respLENG);
		NNTP_CACHED++;
		return fp;
	}
	/*########################################################
	 *	reuse HEAD of previous ARTICLE if possible
	 */
	if( strcmp(command,"HEAD")==0 || strcmp(command,"BODY")==0 ){
		unsigned char acommand[MAX_COMLN],*buf;
		int hc,hl,btop,bc,bl;

		sprintf(acommand,"ARTICLE %ld",number);
		if( strcmp(NNTP_respCOMMAND,acommand) == 0 ){
			buf = NNTP_respBUFF;

			hl = 0;
			btop = 0;
			for( hc = 0; hc < NNTP_respLENG; hc++ ){
				if( buf[hc] == '\n'){
					hl++;
					if( buf[hc+1]=='\n'){
						hc++;
						btop = hc + 1;
						break;
					}
				}
			}
			if( strcmp(command,"HEAD") == 0 ){
				fp = pseudo_fopen(buf,NNTP_respSIZE);
				pseudo_ungetv(fp,hc);
				if(bytes) *bytes = hc;
				if(lines) *lines = hl;
			}else{
				bl = 0;
				for( bc = btop; bc < NNTP_respLENG; bc++ )
					if( buf[bc] == '\n' )
						bl++;

				fp = pseudo_fopen(buf+btop,NNTP_respSIZE-btop);
				pseudo_ungetv(fp,bc-btop);
				if(bytes) *bytes = bc-btop;
				if(lines) *lines = bl;
			}
			nntp_printlog(0,"%s (REUSE-CACHED)",nntp_combuf);
			NNTP_CACHED++;
			return fp;
		}
	}

	/*########################################################
	 *	use prefetched HEAD if possible
	 */
	if( strcmp(NNTP_prefetch_command,nntp_combuf) == 0 ){
		NNTP_HIT++;
		NNTP_prefetch_command[0] = 0;
	}else{
		if( PUT_SERVER(nntp_combuf) != 0 )
			return NULL;
		nntp_cancell();
	}

	/*########################################################
	 *	prefetch next HEAD
	 */
	if( NNTP_DOPREFETCH ){
		if( 2 <= NNTP_contiguous )
		if( strcmp(command,"HEAD") == 0 )
		{
			if( !news_cached(newsgroup,number-1) )
				nntp_prefetch( newsgroup,number-1,command );
			else
			if( !news_cached(newsgroup,number+1) )
				nntp_prefetch( newsgroup,number+1,command );
		}
	}

	fp = nntp_getresponse(nntp_combuf,0,lines,bytes);
	return fp;
}

nntp_pra( command, number )
{	int lines,bytes;
	FILE *fp;
	char line[1024];

	fp = nntp_article(command, NNTP_current_group, number, &lines,&bytes);
	printf("==== %s %s:%d ====\n",command,NNTP_current_group,number);
	while( fgets(line,sizeof(line),fp) != NULL )
		printf("%s",line);
	printf("==== %dlines/%dbytes ====\n",lines,bytes);
	fclose(fp);
}


nntp_midscan(line,mid)
	char *line,*mid;
{	char command[MAX_COMLN],response[MAX_STATLN];

	if( !NNTP_XNNTP_EXISTS && NNTP_WITHOUT_XHIST )
		return 0;

	if( nntp_initialize() < 0 )
		return 0;

	sprintf(command,"XHIST %s",mid);

	if( NNTP_XNNTP_EXISTS ){
		put_server1(NNTP_XNNTP_SERVER,command);
		get_server1(NNTP_XNNTP_SERVER,response,sizeof(response));
		if( *response == CHAR_OK ){
			sscanf(response,"%*d %[^\n]",line);
			dbg_printf("XNNTP: XHIST>%s\n",line);
			return 1;
		}
		strcpy(line," ");
		return 0;
	}

	nntp_cancell();
	if( PUT_SERVER(command) != 0 )
		return 0;

	if( nntp_get_statresp(response,sizeof(response),command) < 0 ){
		if( atoi(response) == ERR_NOART )
			sprintf(line,"?");
		else	*line = 0;
	}else	sscanf(response,"%*d %[^\n]",line);
	dbg_printf("NNTP: XHIST>%s\n",line);
	return 1;
}


static errmessage(f,a1,a2,a3,a4,a5,a6,a7){
	news_hold_message(0);
	message(f,a1,a2,a3,a4,a5,a6,a7);
	news_hold_message(1);
}
static message(f,a1,a2,a3,a4,a5,a6,a7){
	int savstat;

	savstat = NNTP_STAT;
	NNTP_STAT = -1;
	winbar(f,a1,a2,a3,a4,a5,a6,a7);
	NNTP_STAT = savstat;
}

/*
 *
 */
extern nntp_exist();
       int   (*NNTP_EXIST)() = nntp_exist;
extern int   (*NNTP_CONNECT)();
extern char *(*NNTP_SERVER)();
extern FILE *(*NNTP_GETACTIVE)();
extern FILE *(*NNTP_NEWGROUPS)();
extern FILE *(*NNTP_NEWHISTORY)();
extern FILE *(*NNTP_ARTICLE)();
extern int   (*NNTP_MIDSCAN)();
extern int   (*NNTP_FINALIZE)();
extern int   (*NNTP_IDLE)();
extern int   (*NNTP_GETLOG)();

static nntp_exist(){
	if( NNTP_SERVER != nntp_server ){
		/*PREFER_NNTP_HEAD = 1;*/
		NNTP_EXIST = nntp_exist;
		NNTP_CONNECT = nntp_connect;
		NNTP_SERVER = nntp_server;
		NNTP_GETACTIVE = nntp_getactive;
		NNTP_NEWGROUPS = nntp_newgroups;
		NNTP_NEWHISTORY = nntp_newhistory;
		NNTP_ARTICLE = nntp_article;
		NNTP_MIDSCAN = nntp_midscan;
		NNTP_FINALIZE = nntp_finalize;
		NNTP_IDLE = nntp_idle;
		NNTP_GETLOG = nntp_getlog;
	}
	return 1;
}

static DO_WHEN_LOADED(){ nntp_exist(); }
